#include <oxstd.h>
#include <oxprob.h>
#include <oxfloat.h>
#include <oxdraw.h>

#import <maximize>

static decl
	s_ammData,							// array of data
	s_mParameters,						// matrix of initial parameters
	s_ammParameters,					// pointer to array of parameter matrices  over time
	s_mPosteriorMixtureProbabilities,	// matrix iN x iP of posterior mixture probabilities \tau_ij
	s_iP,								// # of mixture components
	s_dFOffset,							// shift of the objective function
	s_mLik;								// matrix iN x iP with component density values


DrawParametric(const iWindow, const ammParameters)
//Draws parameter paths in P=2 dimensions
{
	decl i1,i2,y,x,v,yy,xx,vv;	  //i1: time, i2: cluster

	SetDrawWindow("Time-varying means");
	yy = xx = vv = <>;
	for (i2=0;i2<rows(ammParameters[0]);++i2)
	{
		y = x = v = <>;
		for (i1=0;i1<sizeof(ammParameters);++i1)
		{
			x ~= ammParameters[i1][i2][0];
			y ~= ammParameters[i1][i2][1];
		}
		yy |= y;
		xx |= x;
		DrawXMatrix(iWindow+i2,y,0,x,0,0,2+i2);
	}
	//ShowDrawWindow();
}


DrawVol(const iWindow, const ammParameters)
//Draws vol paths in P=2 dimensions
{
	decl i1,i2,y,x,v,yy,xx,vv;	  //i1: time, i2: cluster

	SetDrawWindow("Time-varying vols");
	for (i2=0;i2<rows(ammParameters[0]);++i2)
	{
		y = x = v = <>;
		for (i1=0;i1<sizeof(ammParameters);++i1)
		{
			v ~= ammParameters[i1][i2][]';
		}
		DrawTMatrix(iWindow+i2, v, "all tv parameters");	
		}

	//ShowDrawWindow();
}



Logistic(const vX) {
//Logistic transformation R -> (0,1)
return ones(vX) ./ (ones(vX)+exp(-vX));	}


LogOdds(const mX) {
//Undoes logistic transformation (0,1) -> R
return log(mX) - log(ones(mX) - mX); }


RestrictPar(const vP) {
//restricts parameters to certain intervals
	decl vPar = vP;
 	vPar[0]   =  2.0* Logistic(vP[0]);
    vPar[1]   =  0.00* Logistic(vP[1]);
	return vPar; }


	
fSimulateData1(const iPlot)
{
// simulate data;
//
// INPUT:
// iPlot: integer; if 1, then plot the data while generating
//
// OUTPUT:

	decl iN = 100, iT = 30, iP = 3, iD = 2;		  //iN = firms per cluster!
	decl vTime = 4*M_2PI * range(1,iT) / iT;
	decl mX, mY, mZ, mAid, mCorr, mChol, iOffset=0;
	decl iWindowScale = 10;

	// initialize
	mX = mY = mZ = zeros(iP*iN,iT);
	s_ammData = new array[iT];

	for (decl i1=0;i1<iT;++i1)
	{

		iOffset = 0;	                             //start with first component!
		s_ammData[i1] = zeros(iP*iN, iD);
	
		// first mixture component
		mCorr = <1,-0.9;-0.9,1>;
		mChol = choleski(mCorr)';
		mAid  = rann(iN,iD) * mChol;	  //iN x 2
		s_ammData[i1][iOffset:iOffset+iN-1][0] = mX[iOffset:iOffset+iN-1][i1] = 4*cos(vTime[i1]) + 0 + mAid[][0];
		s_ammData[i1][iOffset:iOffset+iN-1][1] = mY[iOffset:iOffset+iN-1][i1] = 4*sin(vTime[i1]) + 0 + mAid[][1];
		mZ[iOffset:iOffset+iN-1][i1] = 1;
		if (iPlot==1) DrawXMatrix(0,mY[iOffset:iOffset+iN-1][i1]',0,mX[iOffset:iOffset+iN-1][i1]',0,1,2);
		if (iPlot==1) DrawAdjust(ADJ_SYMBOL,2,150);
		iOffset += iN;
	
		// second mixture component
		mCorr = <0.5,0.2;0.2,0.125>;
		mChol = choleski(mCorr)';
		mAid = rann(iN,iD) * mChol;
		s_ammData[i1][iOffset:iOffset+iN-1][0] = mX[iOffset:iOffset+iN-1][i1] = 3*cos(-vTime[i1]) + 5 + mAid[][0];
		s_ammData[i1][iOffset:iOffset+iN-1][1] = mY[iOffset:iOffset+iN-1][i1] = 3*sin(-vTime[i1]) + 5 + mAid[][1];
		mZ[iOffset:iOffset+iN-1][i1] = 2;
		if (iPlot==1) DrawXMatrix(0,mY[iOffset:iOffset+iN-1][i1]',0,mX[iOffset:iOffset+iN-1][i1]',0,1,3);
		if (iPlot==1) DrawAdjust(ADJ_SYMBOL,2,150);
		iOffset += iN;
	
		// third mixture component
		mCorr = <0.25,0;0,0.25>;
		mChol = choleski(mCorr)';
		mAid = rann(iN,iD) * mChol;
		s_ammData[i1][iOffset:iOffset+iN-1][0] = mX[iOffset:iOffset+iN-1][i1] = 3*sin(vTime[i1])   - 7 + mAid[][0];
		s_ammData[i1][iOffset:iOffset+iN-1][1] = mY[iOffset:iOffset+iN-1][i1] = 3*sin(2*vTime[i1]) - 7 + mAid[][1];
		mZ[iOffset:iOffset+iN-1][i1] = 3;
		if (iPlot==1) DrawXMatrix(0,mY[iOffset:iOffset+iN-1][i1]',0,mX[iOffset:iOffset+iN-1][i1]',0,1,4);
		if (iPlot==1) DrawAdjust(ADJ_SYMBOL,2,150);


		if (iPlot==1) {
			DrawAxis(0,1,0,-iWindowScale,iWindowScale,1,0, 0,0);
			DrawAxis(0,0,0,-iWindowScale,iWindowScale,1,0, 0,0);
			//DrawAxis(0,1,0,-iWindowScale,iWindowScale,-iWindowScale,iWindowScale/10,1,0,TRUE,0,0);
			//DrawAxis(0,0,0,-iWindowScale,iWindowScale,-iWindowScale,iWindowScale/10,1,0,TRUE,0,0);
			DrawAdjust(ADJ_AREA_X,-10,10,1);
			DrawAdjust(ADJ_AREA_Y,-10,10,1);
			ShowDrawWindow();
		}
	}

	s_mPosteriorMixtureProbabilities = ones(iN*iP,iP)/iP;
	
	return {mX, mY, mZ};
	
}



fGasFilter(const ammData, const mPosteriorMixtureProbabilities, mParameters, const vEWMA)
{
// Run the Gaussian GAS filter for the mean and the covariance matrix.
// At the moment, the filter is a single parameter EWMA filter.
//
// INPUTS:
// ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// mPosteriorMixtureProbabilities: iN x iP matrix containing the mixture probabilities 
//				 for firm iN and mixture component iP
// mParameters : iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of cov matrix)'
// dEWMA       : 2x1 double, the GAS smoothing parameter for mean and variance
//
// OUTPUTS:
// ammParameters: iT x iP x (iD + 0.5*iD*(iD+1)) array.
// vELik       :  1 x iP vector holding the conditionally expected log likelihoods for each mixture component
// mLik        :  iN x iP matrix holding the log densities for each mixture component


	decl iD,iN,iT,iP,i1,i2,ix;
	decl ammParameters, vMean, mChol, mCov, vELik, mLik;

	// initialize
	iT = sizeof(ammData);			// number of time points
	iN = rows(ammData[0][][]);	    // number of cross sectional units
	iD = columns(ammData[0][][]);	// number of characteristics in data
	iP = rows(mParameters);         // number of mixture components
	ammParameters = new array[iT];
	vELik = zeros(1,iP);
	mLik = zeros(iN,iP);

	
	// run the filter per time step
	for (i1=0;i1<iT;++i1) {

		// store parameters
		ammParameters[i1] = mParameters;

		// loop over the mixture components
		for (i2=0;i2<iP;++i2) {
		
			// get out the parameters
			ix = 0;
			vMean = mParameters[i2][ix:ix+iD-1];	//vMean refers to component i2 at time i1, of dimension iD x iD
			ix += iD;
			mChol = mParameters[i2][ix:ix+iD*(iD+1)/2-1];
			ix += iD*(iD+1)/2;
			mChol = lower(unvech(mChol));
			mCov = mChol * mChol';

			decl vMixtureProbabilities = meanc(s_mPosteriorMixtureProbabilities);  
			
			// compute the likelihoods
			mLik[][i2] += -0.5*iD*log(M_2PI) - sumr(log(diagonal(mChol)))											// log f_m (x_it | vMean, mCov)
						  -0.5 * sumr(((ammData[i1][][] - vMean) * invertsym(mCov)) .* (ammData[i1][][] - vMean));
			vELik[i2] += mPosteriorMixtureProbabilities[][i2]'mLik[][i2];											//\tau_m' log f_m (x_it | vMean, mCov)
 
			// add the other, second part to the objective function as well.
			vELik[i2] += log(vMixtureProbabilities[i2]) * sumc(mPosteriorMixtureProbabilities[][i2]);
			

			// compute the score steps and update parameters
			vMean = vMean + vEWMA[0] * mPosteriorMixtureProbabilities[][i2]'(ammData[i1][][] - vMean)
			                           /sumc(mPosteriorMixtureProbabilities[][i2]); //posterior mix prob = tau's; computed via Bayes rule. \pi_j is prior, unconditional probability.
		
			mCov  = mCov + vEWMA[1] /sumc(mPosteriorMixtureProbabilities[][i2]) * (
						(ammData[i1][][] - vMean)'(mPosteriorMixtureProbabilities[][i2] .* (ammData[i1][][] - vMean))
						-
						sumc(mPosteriorMixtureProbabilities[][i2]) * mCov );

			// store parameters into matrix object
			ix = 0;
			mParameters[i2][ix:ix+iD-1] = vMean;
			ix += iD;
			mParameters[i2][ix:iD*(iD+1)/2-1] = vech(choleski(mCov))';
			ix += iD*(iD+1)/2;
		}
	}

	//update the initial values for all time-varying GAS mean and cov.
	decl s_mParameters = ammParameters[0][][];
	for (i1=1;i1<iT;++i1) s_mParameters += ammParameters[i1][][];
	s_mParameters /= iT;
	
	return {vELik, mLik, ammParameters};

}



fPerformEStep(const ammData, const mParameters, const vEWMA)
{
// Compute the posterior probabilities, i.e., do the E step of the
// EM algorithm for the mixture.
//
// INPUTS:
// ammData:      iT x iN x iD array, with iN the number of cross-sectional units,
//               and iT the number of time points, and iD the number of characteristics
// amOldPosteriorMixtureProbabilities : pointer to iN x iP matrix containing, on entry containing
//               the posterior mixture probabilities for firm iN and mixture component iP.
//               On exit: containing the updated estimates of this parameter.
// mParameters : iP x (iD + 0.5*iD*(iD+1)) array, with iP the number of mixture components.
//               The columns mParameters[iP][] are
//                [0:iD-1]: means
//                [iD:iD+0.5*iD*(iD+1)-1]: vech(choleski of cov matrix)'
// dEWMA       : 2x1 double, the GAS smoothing parameter for mean and variance
//
// NO OUTPUTS:
// vMixtureProbabilities : 1 x iP vector containing the new mixture probabilities from the M step.
               
	decl vMixtureProbabilities, vWeightedDens, mLik, vELik;

	// run the filter at the latest parameters
	[vELik, mLik, s_ammParameters] = fGasFilter(ammData, s_mPosteriorMixtureProbabilities, mParameters, vEWMA);

	// compute new prior probabilities \pi from old \tau_ij's
	vMixtureProbabilities = meanc(s_mPosteriorMixtureProbabilities);   		 //eq 7 scribbles

	//save matrix with component log density values
	s_mLik = mLik;

	// compute the new posterior probabilities.
	// NOT ALWAYS FEASIBLE:
	vWeightedDens = (vMixtureProbabilities .* exp(mLik - maxr(mLik)));
	s_mPosteriorMixtureProbabilities = vWeightedDens ./ sumr(vWeightedDens);

	return 1; //{vMixtureProbabilities};

}




fMaximizeStep1(const vP, const adF, const avG, const amH)
{
// Perform the maximization step and return the conditionally expected
// log likelihood function.
// This routine has the input/output format suited for the MaxBFGS optimizer.
//
// Data: should be known in the array iT x iN x iD variable s_ammData
//
// THIS VERSION INPUT vP:
// estimate the single smoothing parameteter for the mean,
// the initial means, and
// all of the mixture component covariance matrices matrices 

	decl ix,iP,iN,iD,iT;
	decl vMixtureProbabilities, dEWMA;

	// get the dimensions
	iT = sizeof(s_ammData);
	iN = rows(s_ammData[0]);
	iD = columns(s_ammData[0]);
	iP = s_iP;
	if (fmod(iP,1) != 0) {
		println("*** ERROR in fMaximizeStep1(..): dimensions of vP[",rows(vP),"] incorrect ***");
		println("iT = ",iT,"\niN = ",iN,"\niD = ",iD,"\niP = ",iP);
		exit(0);
	}
	
	decl mParameters = s_mParameters; 
	//restrict parameters
	decl vPar =  RestrictPar(vP);

	// run the filter
	decl ELik, Lik, ammP;
	[ELik, Lik, ammP] = fGasFilter(s_ammData, s_mPosteriorMixtureProbabilities, mParameters, vPar);
	DrawParametric(0,ammP);  
	DrawVol(4,ammP);
	ShowDrawWindow();

//	println(ammP);
	adF[0] = sumr(ELik) - s_dFOffset;
	//println("dF:", adF[0]);
	if (ismissing(adF[0])) return 0; else return 1;
}



//fEmExpectationStepTimeT(const aData, const aParameters) {
//// do the mixture expectation step
////
//// INPUTS:
//// vData        : iN x 2 vector of data points
//// mParametersk : iM x 6 matrix, with iM the number of mixture components
////                Columns:
////                [0]: mixture probability
////                [1:2]: means
////                [3:5]: vech(cov matrix)
////
//// OUTPUTS:
//
//	decl vLiknt, dProb, vMean, mChol, mCov;
//	decl i1;
//
//	// initialize
//	vLiknt = zeros(rows(vData) , rows(mParametersk));
//
//	// compute likelihoods for each regime for all observations
//	for (i1=0;i1<rows(mParametersk);++i1) {
//
//		dProb = 1.0 / (1 + exp(-mParametersk[i1][0]));
//		vMean = mParametersk[i1][1:2];
//		mChol = (mParametersk[i1][3] ~ 0) | (mParametersk[i1][4] ~ mParametersk[i1][5]);
//		mCov = mChol*mChol';
//		
//		vLiknt[][i1] = dProb * 
//			(
//				exp(-0.5 * sumr(((vData - vMean) * invertsym(mCov)) .* (vData - vMean)))
//				/
//				(M_2PI * mChol[0][0] * mChol[1][1])
//			);
//	}
//
//	return { vLiknt ./ sumr(vLiknt), vLiknt};
//
//}




main() {

	println("simulating");
	decl mX, mY, mZ;
	[mX, mY, mZ] = fSimulateData1(1);
	s_mPosteriorMixtureProbabilities = <0.6,0.2,0.2;0.2,0.6,0.2;0.2,0.2,0.6>[mZ[][0]-1][];
	s_mPosteriorMixtureProbabilities = <0.2,0.6,0.2;0.2,0.6,0.2;0.2,0.6,0.2>[mZ[][0]-1][];
	
	println("evaluating"); 
//	decl mParameters = <7,0,1,-0.9,0.43;9.5,-1,0.7,0,0.35;0,0,0.5,0,0.5>;
	s_mParameters =  0 ~ 0 ~ vech(choleski(<1   ,-0.9;-0.9, 1    >))' |		//initial parameters for mean and covariance matrix.	 
	  				 5 ~ 5 ~ vech(choleski(<0.5 ,0.2 ;0.2 , 0.125>))' | 
					-7 ~-7 ~ vech(choleski(<0.25,0   ;0   , 0.25 >))' ;

	decl dEWMAmean = -0.27;
	decl dEWMAvar  = 0;
	decl vP, vPar, dF=0;
	vPar = vP = dEWMAmean | dEWMAvar;
	s_iP = 3;
	
	//first M-Step to initialize \tau_ij's: 
	s_dFOffset = 0;
//	fMaximizeStep1(vP, &dF, 0, 0);
//	println("\nDone the initial M step: got tv means and variances"); 
	
	//Iterate E and M steps:
	decl i, ir, dFold=1, dconv = 1;
	MaxControl(-1, 1, 1);
	MaxControlEps(1e-4, 1e-3);

	for (i = 0; i < 4; i++)
	{
		//E-Step, producing s_mPosteriorMixtureProbabilities 
		println("E-Step"+sprint(i));
		savemat("results/s_mPosteriorMixtureProbabilities_v"+sprint(i)+".in7", s_mPosteriorMixtureProbabilities); 
		savemat("results/s_mLik_v"+sprint(i)+".in7", s_mLik);
		
		fPerformEStep(s_ammData, s_mParameters, RestrictPar(vP));

		//M-Step
		println("M-Step"+sprint(i));

		ir = MaxBFGS(fMaximizeStep1, &vP, &dF, 0, 1);	  
		
		print("vP unrestricted, restricted: ", vP ~ RestrictPar(vP));
		
 		if (i>2) dconv = max(fabs((dF - dFold) ./ dFold));
		if (dconv < 10^-7) break;
		dFold = dF;
		
	};

	print("vP unrestricted: ", vP);
	print("vPar restricted: ", RestrictPar(vP)); 
	
}